#pragma once

#include <type_traits>

#include <dune/common/hybridutilities.hh>
#include <dune/functions/functionspacebases/boundarydofs.hh>
#include <dune/functions/functionspacebases/interpolate.hh>
#include <dune/functions/functionspacebases/subspacebasis.hh>

#include <amdis/LinearAlgebra.hpp>

namespace AMDiS {

template <class D, class R>
  template <class RB, class CB>
void DirichletBC<D,R>::
init(RB const& rowBasis, CB const& colBasis)
{
  using Dune::Functions::forEachBoundaryDOF;
  using LV = typename CB::LocalView;
  using IS = typename CB::GridView::Intersection;
  dirichletNodes_.resize(colBasis.dimension());
  forEachBoundaryDOF(colBasis, [&](int localIndex, LV const& localView, IS const& intersection) {
    dirichletNodes_[localView.index(localIndex)] = onBoundary(intersection);
  });
}


template <class D, class R>
  template <class Mat, class Sol, class Rhs, class RN, class RTP, class CN, class CTP>
void DirichletBC<D,R>::
fillBoundaryCondition(Mat& matrix, Sol& solution, Rhs& rhs, RN const& rowNode, RTP rowTreePath, CN const& colNode, CTP colTreePath)
{
  auto columns = Constraints<Mat>::dirichletBC(matrix, dirichletNodes_);

  Dune::Hybrid::ifElse(std::is_same<RangeType_t<RN>, R>{},
    [&](auto id) {
      auto subBasis = Dune::Functions::subspaceBasis(*rhs.basis(), rowTreePath);
      Dune::Functions::interpolate(subBasis, rhs, values_, dirichletNodes_);
    });

  // subtract columns of dirichlet nodes from rhs
  for (auto const& triplet : columns)
    rhs[triplet.row] -= triplet.value * solution[triplet.col];

  Dune::Hybrid::ifElse(std::is_same<RangeType_t<CN>, R>{},
    [&](auto id) {
      auto subBasis = Dune::Functions::subspaceBasis(*solution.basis(), colTreePath);
      Dune::Functions::interpolate(subBasis, solution, values_, dirichletNodes_);
    });
}

} // end namespace AMDiS
